/* 
   see main() for syntax summary.

   scans an rpg file, updating the master palette to 
   new colors, and updating graphics accordingly.

   modifies the following lumps:

   .PAL .MXS .TIL .MAS

   updating is done in-place; there is no unpacking of the rpgfile.

*/

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>

#define ABS(a)     (((a) < 0) ? -(a) : (a))
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))

typedef struct RGB {

	unsigned char r,g,b;
	unsigned char pad;
} RGB;

unsigned int maptable[256];

RGB old_palette[256];
RGB new_palette[256];

/* 1.5k lookup table for color matching */
static unsigned int col_diff[3*128]; 

static unsigned char dither_table[7] = { 16, 68, 146, 170, 109, 187, 239 };
static unsigned char dither_ytable[8] = { 1, 5, 2, 7, 4, 0, 6, 3 };


int read_size(FILE *i)
{
	int hb, _hb, _lb, lb;

	hb = fgetc(i);
	_hb = fgetc(i);
	_lb = fgetc(i);
	lb = fgetc(i);

	return ( _lb | (lb << 8) | (hb << 16) | (_hb << 24) );


}

void write_size(int size, FILE *o)
{
	int hb, _hb, _lb, lb;

	hb = (size >> 16) & 0xff;
	_hb = (size >> 24) & 0xff;
	_lb = size & 0xff;
	lb = (size >> 8) & 0xff;

	fputc( hb,o);
	fputc( _hb,o);
	fputc( _lb,o);
	fputc( lb,o);
}

/* bestfit_init and bestfit_color are from allegro, modified */

/* bestfit_init:
 *  Color matching is done with weighted squares, which are much faster
 *  if we pregenerate a little lookup table...
 */
static void bestfit_init(void)
{
   int i;

   for (i=1; i<64; i++) {
      int k = i * i;
      // These weights approximate Euclidean distance in the YUV space
      // See https://www.compuphase.com/cmetric.htm
      // (The weights previously used, from Allegro, were nonsensical)
      // -- TMC170529
      col_diff[0  +i] = col_diff[0  +128-i] = k * 4;
      col_diff[128+i] = col_diff[128+128-i] = k * 3;
      col_diff[256+i] = col_diff[256+128-i] = k * 2;
   }
}



/* bestfit_color:
 *  Searches a palette for the color closest to the requested R, G, B value.
 */
int bestfit_color(const RGB *p, int r, int g, int b)
{
   int i, coldiff, lowest, bestfit;

   if (col_diff[1] == 0)
      bestfit_init();

   bestfit = 0;
   lowest = INT_MAX;

   i = 0;

   while (i<256) {
      const RGB *rgb = &p[i];
      coldiff = (col_diff + 0) [ (rgb->g - g) & 0x7F ];
      if (coldiff < lowest) {
	 coldiff += (col_diff + 128) [ (rgb->r - r) & 0x7F ];
	 if (coldiff < lowest) {
	    coldiff += (col_diff + 256) [ (rgb->b - b) & 0x7F ];
	    if (coldiff < lowest) {
	       bestfit = rgb - p;    /* faster than `bestfit = i;' */
	       if (coldiff == 0)
		  return bestfit;
	       lowest = coldiff;
	    }
	 }
      }
      i++;
   }

   return bestfit;
}


/* makecol_dither:
 *  Calculates a dithered 8 bit pixel value.
 *  adapted from allegro makecol16_dither.
 */
int makecol_dither(int r, int g, int b, int x, int y)
{
   int returned_r, returned_g, returned_b;
   int bpos;

   returned_r = r/8;
   returned_b = b/8;
   returned_g = g/8;

   y = dither_ytable[y&7];

   if (r&7) {
      bpos = (x+y)&7; 
      returned_r += ((dither_table[((r&7)-1)]) & (1<<bpos)) >> bpos;
   }

   if (b&7) {
      bpos = (x+y+3)&7; 
      returned_b += ((dither_table[((b&7)-1)]) & (1<<bpos)) >> bpos;
   }

   if (g&7) {
      bpos = (x+y+2)&7;
      returned_g += ((dither_table[(g&7)-1]) & (1<<bpos)) >> bpos;
   }

   if (returned_r > 31)
      returned_r = 31;

   if (returned_b > 31) 
      returned_b = 31;

   if (returned_g > 31) 
      returned_g = 31;

   return bestfit_color(new_palette,returned_r << 1, returned_g << 1, returned_b << 1);

}

int extension_matches(char *filename, char *extension)
{
	int l = strlen(filename);

	if (filename[l-1] == extension[2] && filename[l-2] == extension[1] && filename[l-3] == extension[0])
		return 1;

	return 0;
}

void copy_bytes(FILE *i, FILE *o, int length)
{
	int j;

	for (j = 0; j < length && !feof(i); j++)
		fputc(fgetc(i),o);
}

void translate_with_dithering(FILE *i,FILE *o,int length)
{
	int j;
	int x,y;
	int t;

	for (j = 0; j < length; j ++)
	{
		y = (j / 80) % 200;
		x = (j % 80) * 4 + ( (j % 64000) / 16000);

		t = fgetc(i);
		t = makecol_dither(old_palette[t].r << 2, old_palette[t].g << 2, old_palette[t].b << 2, x, y);

		fputc(t,o);
	}


}

void translate_indices (FILE *i,FILE *o,int length)
{
	int j;

	for (j = 0; j < length; j ++)
	{
		fputc(maptable[fgetc(i)],o);
	}
}

void mxs_and_til (FILE *i,FILE *o,int length, int dither)
{
	printf("translating mxs/til %d\n", length);

	if (dither)
		translate_with_dithering(i,o,length);
	else
		translate_indices(i,o,length);
}

void pal (FILE *i, FILE *o, int length)
{
	int j;

	printf("translating pal %d\n" , length);

	/* ignore header */
	for (j = 0; j < 7;j ++)
		fputc(fgetc(i),o);

	length -=7;
	translate_indices(i,o,length);

}

/* this gets called on the second iteration, not the first.
   the first reads only */
void mas (FILE *i, FILE *o, int length)
{
	int j;

	printf("translating mas %d\n",length);

	/* ignore header */
	for (j = 0; j < 7;j ++)
		fputc(fgetc(i),o);
	length -=7;

	for (j = 0;j < 256;j ++)
	{
		fgetc(i);
		fgetc(i);
		fgetc(i);
		fgetc(i);
		fgetc(i);
		fgetc(i);
		fputc(new_palette[j].r,o);
		fputc(0,o);
		fputc(new_palette[j].g,o);
		fputc(0,o);
		fputc(new_palette[j].b,o);
		fputc(0,o);
	}

	length -= 1536;

	/* commonly some unused bits at the end */
	for (j = 0;j < length;j ++)
	{
		fputc(fgetc(i),o);
	}
}

void read_mas(FILE *i, int length, RGB *p)
{
	int j;

	printf("reading mas %d\n", length);

	/* ignore header */
	for (j = 0;j < 7;j ++)
		fgetc(i);

	length -=7;

	for (j = 0;j < 256;j ++)
	{
		p[j].r = fgetc(i);
		fgetc(i);
		p[j].g = fgetc(i);
		fgetc(i);
		p[j].b = fgetc(i);
		fgetc(i);
	}

	length -= 1536;

	for (j = 0;j < length;j ++)
		fgetc(i);
}

void build_maptable()
{
	int i;

	for (i = 0; i < 256;i++)
	{
		maptable[i] = bestfit_color(new_palette, old_palette[i].r, old_palette[i].g, old_palette[i].b);
	}

}

void read_old_palette(RGB *p, FILE *i)
{
	char buffer[51];
        buffer[50] = 0;

	int size,t,j;

	while (!feof(i))
	{
		t = 1;

		for (j = 0; t != 0; j++)
		{
			if (j == 50)
                        {
                            printf("ERROR: bad RPG file (lump name too long)\n");
                            return;
                        }
			t = fgetc(i);
			buffer[j] = toupper(t);
			//printf("length %d..\n",j);
		}

		size = read_size(i);

		if (extension_matches(buffer,"MAS"))
		{ 
			puts("reading old palette");
			read_mas(i, size, old_palette);
			return;
		}
		else
		{
			printf("skipping %s (%d bytes)\n", buffer, size);
			for (j = 0; j < size; j++)
				fgetc(i);
		}
	}
}

void remap_rpg_file_pixels(FILE *i, FILE *o)
{
	char buffer[51];
        buffer[50] = 0;

	int j,t;
	int size;

	build_maptable();

	while (!feof(i))
	{
		t = 1;

		for (j = 0; t != 0 && !feof(i); j++)
		{
                    	if (j == 50)
                        {
                            printf("ERROR: bad RPG file (lump name too long)\n");
                            return;
                        }

			if ( feof(i) ) 
				return;

			t = fgetc(i);
			if ( feof(i) ) 
				return;
			fputc(t,o);
			buffer[j] = toupper(t);
		}

		size = read_size(i);
		write_size(size,o);

		if (extension_matches(buffer,"MAS"))
			mas(i,o,size);
		else if (extension_matches(buffer,"MXS") || extension_matches(buffer,"TIL") )
			mxs_and_til(i,o,size,0);
		else if (extension_matches(buffer,"PAL") )
			pal(i,o,size);
		else
		{
			printf("copying %s (%d bytes) \n",buffer, size);
			copy_bytes(i,o,size);
			if (feof(i))
				return;
		}

	}

}

int main(int argc, char *argv[])
{
	FILE *i;
	FILE *o;

	if (argc < 4 || strcmp(argv[1],argv[3]) == 0 )
	{
		printf("syntax: %s <rpgfile.rpg> <newpalette.mas> <outfile.rpg>\n\n",argv[0]);
		printf("input and output filenames must be different. if necessary, change the input file's extension to something like .old or .bak.");
		return 0;
	}

	i = fopen(argv[2],"rb");

	if (!i)
	{
	    printf("ERROR: couldn't open MAS file %s\n",argv[2]);
	    return -1;
	}

	read_mas(i,1536 + 7, new_palette);

	fclose(i);

	i = fopen(argv[1],"rb");
	if (!i)
	{
	    printf("ERROR: couldn't open input file %s\n",argv[1]);
	    return -1;
	}

	read_old_palette(old_palette,i);
	fclose(i);

	i = fopen(argv[1],"rb");
	o = fopen(argv[3],"wb");

	if (!i)
	{
	    printf("ERROR: couldn't open output file %s\n",argv[1]);
	    return -1;
	}

	remap_rpg_file_pixels(i,o);
	fclose(i);
	fclose(o);

	return 0;
}
